import filecmp
import os
import os.path
import platform
import sys
import difflib

# import subprocess
# import numpy as np
# import pandas as pd


globalIgnoreLines = {}


class FilteredFileReader:
    """
    a file reader that optionally selects lines containing a substring
    """

    def __init__(self, file_path, filter_tag=None, skip_lines=None):
        """

        :param file_path:
        :param filter_tag: keep only lines containing that substring
        :param skip_lines: list of line indexes to keep
        """
        self._file = open(file_path, 'r')
        self._filter_tag = filter_tag
        self._skip_lines = skip_lines

    def readlines(self):

        if self._filter_tag is None:
            lines = self._file.readlines()
        else:
            lines = [line.split(self._filter_tag)[-1] for line in self._file.readlines()]

        if self._skip_lines is None:
            return lines
        else:
            return [line for i, line in enumerate(lines) if i not in self._skip_lines]


def default_file_diff(file_path1, file_path2,
                      print_diff=True,
                      ignore_patterns=tuple(),
                      filter_tag=None,
                      skip_lines=None,
                      max_diff_in_log=25):
    """ compare two files line by line:
        @:return: False if a difference is found
    """

    try:
        filter_tag1, filter_tag2 = filter_tag
    except:
        filter_tag1 = filter_tag2 = filter_tag

    file1 = FilteredFileReader(file_path1, filter_tag=filter_tag1, skip_lines=skip_lines)
    file2 = FilteredFileReader(file_path2, filter_tag=filter_tag2, skip_lines=skip_lines)

    try:
        diff = difflib.unified_diff(file1.readlines(), file2.readlines(),
                                    # fromfile=file_path1, tofile=file_path2,
                                    n=0  # no context lines
                                    )

        num_diffs = 0
        ignored_diffs = 0
        first_diff = True

        for diff_line in diff:

            if first_diff:
                first_diff = False
                print()

            if any(diff_line.startswith(tag) for tag in ['---', '+++', '@@ ']):
                # skip diff output meta data
                continue
            else:

                ignore = False
                for ignore_pattern in ignore_patterns:
                    # assume it's a list of patterns
                    if callable(ignore_pattern):
                        ignore = ignore_pattern(diff_line)
                    else:
                        ignore = ignore_pattern in diff_line

                    if ignore:
                        ignored_diffs += 1
                        break

                if ignore:
                    continue

            if print_diff:
                print(diff_line, end='', file=sys.stderr)

            num_diffs += 1

            if max_diff_in_log == 0:
                return False

            if num_diffs >= max_diff_in_log:
                print(f'Only showing {max_diff_in_log} differences out of {num_diffs} '
                      f'- diff file manually for more details', file=sys.stderr)
                break

        if num_diffs == 0:
            if ignored_diffs > 0 and platform.system() != 'Windows':
                print('        Ignoring Diff on {0} coming from line endings unix vs windows'.format(file_path1))

            return True

        return False

    except UnicodeDecodeError:
        print("     skipping text diff on non-text file:", file_path1)
        return False


def diff_dirs(ref_dir, test_dir, print_diff=True, ignore_lines=None, comparators=None, ignore_missing=False):
    """
    recursive directory compare:
    :param ref_dir:
    :param test_dir:
    :param print_diff:
    :param ignore_lines: {file_ext: [ string1, string2, ... ]}
        where the differing line will be ignored if it contains any of the strings
    :param comparators: {file_ext: comparator_func}
        where comparator_func: f(file1, file2) -> bool (True if no diff)
    :param ignore_missing:
    :return: differing files in @ref_dir vs @test_dir
    """

    dir_comp_res = filecmp.dircmp(ref_dir, test_dir)

    # print out what was found in each directory for information
    if print_diff:
        print()
        dir_comp_res.report()
        print()

    if not ignore_missing:
        # check number of files is the same
        assert len(dir_comp_res.right_only) == 0, f'Files only in {dir_comp_res.right}: {dir_comp_res.right_only}'
        assert len(dir_comp_res.left_only) == 0, f'Files only in {dir_comp_res.left}: {dir_comp_res.left_only}'

        # check names of files are the same
        assert set(dir_comp_res.right_list).issuperset(dir_comp_res.left_list), \
            f'Different file names found {set(dir_comp_res.right_list).difference(dir_comp_res.left_list)}'

    for diff_file in dir_comp_res.diff_files:

        file_name1 = os.path.join(ref_dir, diff_file)
        file_name2 = os.path.join(test_dir, diff_file)

        if ignore_missing:
            if not os.path.exists(file_name1):
                missing = file_name1
            elif not os.path.exists(file_name2):
                missing = file_name2
            else:
                missing = None

            if missing:
                print(f'missing directory', missing, file=sys.stderr)
                continue

        found_comparator = False
        # use a custom comparator fun if it exists for that file extension
        for comparator_dict in [comparators, globalIgnoreLines]:
            for file_ext in comparator_dict:
                if diff_file.endswith(file_ext):
                    found_comparator = True

                    if comparator_dict[file_ext](file_name1, file_name2):
                        yield diff_file

                    break

        if not found_comparator:
            if default_file_diff(file_name1, file_name2, print_diff, ignore_lines):
                yield diff_file

    for file_name in os.listdir(ref_dir):
        if os.path.isdir(os.path.join(ref_dir, file_name)):
            dir_path1 = os.path.join(ref_dir, file_name)
            dir_path2 = os.path.join(test_dir, file_name)

            if ignore_missing:
                if not os.path.exists(dir_path1):
                    missing = dir_path1
                elif not os.path.exists(dir_path2):
                    missing = dir_path2
                else:
                    missing = None

                if missing:
                    print('Missing directory', missing, file=sys.stderr)
                    continue

            for diff_file in diff_dirs(dir_path1, dir_path2, print_diff=print_diff,
                                       ignore_lines=ignore_lines,
                                       comparators=comparators,
                                       ignore_missing=ignore_missing):
                yield os.path.join(file_name, diff_file)


class FileComparator:

    def __init__(self, include_line_if_contains=None, skip_lines=None, **kwargs):

        self.include_line_if_contains = include_line_if_contains
        self.skip_lines = skip_lines
        self.kwargs = kwargs

    def description(self):
        if self.include_line_if_contains:
            if isinstance(self.include_line_if_contains, str):
                include_str = f'"{self.include_line_if_contains}"'
            else:
                include_str = ", ".join(f'"{line}"' for line in self.include_line_if_contains if line is not None)
            return f'filter lines on {include_str}'
        return ''

    def __call__(self, *args, **kwargs):
        return default_file_diff(*args, **kwargs,
                                 filter_tag=self.include_line_if_contains,
                                 skip_lines=self.skip_lines,
                                 **self.kwargs)



import os
import sys
import shutil
import time
from contextlib import contextmanager
import subprocess
import typing
import pathlib


def print_test_meta(*message):
    print(*message, file=sys.stderr)


def rmdir(dir_path):
    if os.path.exists(os.path.join(os.getcwd(), dir_path)):
        shutil.rmtree(dir_path)


@contextmanager
def working_dir(path):
    """temporary switch of working directory"""

    current_dir = os.getcwd()
    if path is not None:
        os.chdir(path)

    try:
        yield
    finally:
        if path is not None:
            os.chdir(current_dir)


def prompt_user(message_prompt: str,
                yes_entries=('y',),
                no_entries=('n',),
                default: bool = None) -> bool:
    if default is not None:
        if default:
            reply_usage = f'[{yes_entries[0]}]/{no_entries[0]}'
        else:
            reply_usage = f'{yes_entries[0]}/[{no_entries[0]}]'
    else:
        reply_usage = f'{yes_entries[0]}/{no_entries[0]}'

    message_prompt += ' ' + reply_usage + '\n'

    while True:
        user_input = input(message_prompt)
        if user_input == '' and default is not None:
            return default
        if user_input in yes_entries:
            return True
        if user_input in no_entries:
            return False

        print(f'Invalid choice: {user_input}')


@contextmanager
def revert_file_modif(file_path):
    """ Saves file content and  write original content back on exit.
        Creates file if it does not exist and deletes it on exist.
    """

    file_exists = os.path.exists(file_path)
    with open(file_path) as f:
        original_content = f.read()

    try:
        yield original_content
    finally:
        if not file_exists:
            os.remove(file_path)
        else:
            with open(file_path, 'w') as f:
                f.write(original_content)


@contextmanager
def perf_timer(message='Time elapsed:', verbose=True):
    """timer context (includes time spent during process sleep)"""
    start_time = time.perf_counter()
    try:
        yield
    finally:
        if verbose:
            time_elapsed = time.perf_counter() - start_time
            print(message, '%.3fsec' % time_elapsed)


def diff_directory(path1, path2):
    """
    Iterate on dir tree below path2 and report if  a path under path2 does not exist under path2
    If some file or dir should be ignored, add them to ignored_dir_or_file.
    :param path1:
    :param path2:
    :return:
    """
    ignored_dir_or_file = ['.svn', '.git', '.gitignore', 'README.md']

    for path in os.listdir(path2):
        if path not in ignored_dir_or_file:
            abs1 = os.path.join(path1, path)
            abs2 = os.path.join(path2, path)

            if not os.path.exists(abs1):
                yield abs2

            if os.path.isdir(abs2):
                for diff_path in diff_directory(abs1, abs2):
                    yield diff_path


class ExeFailedException(Exception):
    pass


def format_log_for_exception(exc, max_num_lines=20):
    """
    extract information from streams captured by a subprocess error
    :param exc: an instance of subprocess.CalledProcessError
    :param max_num_lines: max number of log lines to append to reraised exception
    :return: formatted exception message with extra info from the captured executable log
    """

    exec_log = exc.stderr
    if not exec_log:
        exec_log = exc.stdout

    if exec_log:
        message = exec_log.decode('latin-1', errors='ignore')
        last_few_lines = os.linesep.join(message.rsplit(os.linesep, max_num_lines)[-(max_num_lines-1):])
        msg_len = len(message)

        if msg_len > len(last_few_lines):
            return last_few_lines + f'\n[truncated to last {max_num_lines} lines\n'
        else:
            return last_few_lines


def handle_subprocess_error(exc):
    message = format_log_for_exception(exc)
    if message:
        raise ExeFailedException(message) from exc


def exec_cmdline(command, args_list, check_ret_code=True,
                 log_save_path=None, env_vars=None,
                 pre_cmd=None, post_cmd=None, verbose=True, norun=False):
    """
    Runs command and returns the captured stdout log
    :param command:
    :param args_list:
    :param check_ret_code:
    :param log_save_path:
    :param env_vars:
    :param pre_cmd:
    :param post_cmd:
    :param verbose:
    :param norun:
    :return:
    """

    env_vars_line = ''
    if env_vars:
        env_vars_line = ' '.join([k + '=' + str(v) for k, v in env_vars.items()])
        env_vars_line += ' '

    if isinstance(args_list, str):
        args_list = [args_list]

    # get rid of new lines in command line args
    args_list = [' '.join(line.strip() for line in args.splitlines()) for args in args_list]

    cmd_line_tasks = [command + ' ' + args for args in args_list]

    # chain commands
    cmd_line = ''
    for task in cmd_line_tasks[:-1]:
        cmd_line += '{ ' + env_vars_line + task + ' & } ; '

    cmd_line += env_vars_line + cmd_line_tasks[-1]

    # add pre/post commands
    if pre_cmd:
        cmd_line = ' && '.join(pre_cmd) + ' && ' + cmd_line

    if post_cmd:
        cmd_line += ' && ' + ' && '.join(post_cmd)

    if verbose:
        print()
        print('working dir/command line:')
        print(os.getcwd())
        print(cmd_line)
        print()

    if not norun:
        if not os.path.exists(command):
            raise Exception(f'{command} '
                            f'not found from {os.getcwd()}. '
                            f'Is it in your path? Did you compile?')

        exe_name = os.path.basename(command.split()[0])
        try:
            with perf_timer(exe_name + ' execution took '):
                result_info = subprocess.run(cmd_line, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

            log_out = result_info.stdout
            if log_save_path is not None:
                # log_out = log_out.decode('latin-1')
                with open(log_save_path, 'wb') as f:
                    f.write(log_out)

            if check_ret_code:
                result_info.check_returncode()

            return log_out

        except subprocess.CalledProcessError as exc:
            handle_subprocess_error(exc)
            raise

    return cmd_line


def pattern_matches(pattern: typing.Union[str, typing.List[str]],
                    path_string: str) -> bool:

    if isinstance(pattern, str):
        patterns = [pattern]
    else:
        patterns = pattern

    has_suppress = False
    has_keep = False
    keep = None

    def match(pattern, string):
        if '*' in pattern:
            return pathlib.PurePath(string).match(pattern)
        else:
            return pattern in string

    for pattern in patterns:
        if pattern.startswith('~'):
            has_suppress = True
            if match(pattern=pattern[1:], string=path_string):
                keep = False

        else:
            has_keep = True
            if match(pattern=pattern, string=path_string):
                keep = True

    if keep is None:
        # default
        keep = has_suppress and not has_keep

    return keep
